/*
 * Copyright © 2021-2023 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of ipspiel24-tictactoe-core.
 *
 * ipspiel24-tictactoe-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * ipspiel24-tictactoe-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * ipspiel24-tictactoe-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.ipspiel24.tictactoe.core.domain;

import java.util.List;
import java.util.Map;
import java.util.Set;

import de.fhdw.gaming.core.domain.Stateful;

/**
 * Represents the Tic Tac Toe board.
 * <p>
 * A Tic Tac Toe board is a square and hence possesses an identical non-zero positive number of rows and columns. Each
 * combination of a row and a column is described by a {@link TicTacToePosition position}. At each position, there is a
 * {@link TicTacToeField field} which describes whether the field is empty or marked by a cross or nought.
 */
public interface TicTacToeBoard extends Stateful {

    /**
     * Returns the number of rows (or columns) of this board.
     */
    int getSize();

    /**
     * Checks whether this board contains a field at the position passed.
     *
     * @param position The position.
     * @return {@code true} if this board contains a field at the position passed, else {@code false}.
     */
    boolean hasFieldAt(TicTacToePosition position);

    /**
     * Returns the field at the given position.
     *
     * @param position The position of the field to return.
     * @return The field.
     * @throws IllegalArgumentException if the position is out of range, i.e. if it does not denote a field.
     */
    TicTacToeField getFieldAt(TicTacToePosition position);

    /**
     * Returns all fields of this board line by line.
     * <p>
     * The list(s) returned are possibly immutable, so do not modify them.
     */
    List<List<? extends TicTacToeField>> getFields();

    /**
     * Returns all fields of a given state.
     * <p>
     * The map returned returned is possibly immutable, so do not modify it.
     *
     * @param fieldState The state of the fields to return.
     * @return The fields.
     */
    Map<TicTacToePosition, ? extends TicTacToeField> getFieldsBeing(TicTacToeFieldState fieldState);

    /**
     * Returns all uniformly marked rows on the board. Empty rows are not returned.
     * <p>
     * The set returned returned is possibly immutable, so do not modify it.
     */
    Set<TicTacToeRow> getRowsUniformlyMarked();

    @Override
    TicTacToeBoard deepCopy();
}
