/*
 * Copyright © 2021-2023 Fachhochschule für die Wirtschaft (FHDW) Hannover
 *
 * This file is part of ipspiel24-tictactoe-core.
 *
 * ipspiel24-tictactoe-core is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * ipspiel24-tictactoe-core is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * ipspiel24-tictactoe-core. If not, see <http://www.gnu.org/licenses/>.
 */
package de.fhdw.gaming.ipspiel24.tictactoe.core.domain.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;

import de.fhdw.gaming.ipspiel24.tictactoe.core.domain.TicTacToeBoard;
import de.fhdw.gaming.ipspiel24.tictactoe.core.domain.TicTacToeDirection;
import de.fhdw.gaming.ipspiel24.tictactoe.core.domain.TicTacToeField;
import de.fhdw.gaming.ipspiel24.tictactoe.core.domain.TicTacToeFieldState;
import de.fhdw.gaming.ipspiel24.tictactoe.core.domain.TicTacToePosition;
import de.fhdw.gaming.ipspiel24.tictactoe.core.domain.TicTacToeRow;
import de.fhdw.gaming.ipspiel24.tictactoe.core.domain.TicTacToeRowType;

/**
 * Implements {@link TicTacToeBoard}.
 */
final class TicTacToeBoardImpl implements TicTacToeBoard {

    /**
     * The minimum number of rows (and columns).
     */
    private static final int MINIMUM_BOARD_SIZE = 2;

    /**
     * The fields of the board, sorted first by row and then by column.
     */
    private final Map<Integer, Map<Integer, TicTacToeFieldImpl>> fields;

    /**
     * Creates a Tic Tac Toe board.
     *
     * @param size The number of rows (and columns). It must be a positive number greater than or equal to 2.
     * @throws IllegalArgumentException if the size does not meet the requirements.
     */
    TicTacToeBoardImpl(final int size) throws IllegalArgumentException {
        if (size < TicTacToeBoardImpl.MINIMUM_BOARD_SIZE) {
            throw new IllegalArgumentException(
                    String.format(
                            "The board size %d is not a positive number >= %d.",
                            size,
                            TicTacToeBoardImpl.MINIMUM_BOARD_SIZE));
        }

        this.fields = new LinkedHashMap<>();
        for (int rowIndex = 0; rowIndex < size; ++rowIndex) {
            final Map<Integer, TicTacToeFieldImpl> row = new HashMap<>();
            this.fields.put(rowIndex, row);

            for (int columnIndex = 0; columnIndex < size; ++columnIndex) {
                row.put(
                        columnIndex,
                        new TicTacToeFieldImpl(
                                this,
                                TicTacToePosition.of(rowIndex, columnIndex),
                                TicTacToeFieldState.EMPTY));
            }
        }
    }

    /**
     * Copies an TicTacToe board.
     *
     * @param source The board to copy.
     */
    private TicTacToeBoardImpl(final TicTacToeBoardImpl source) {
        Objects.requireNonNull(source, "source");

        final int size = source.getSize();
        this.fields = new LinkedHashMap<>();
        for (int rowIndex = 0; rowIndex < size; ++rowIndex) {
            final Map<Integer, TicTacToeFieldImpl> row = new HashMap<>();
            this.fields.put(rowIndex, row);

            for (int columnIndex = 0; columnIndex < size; ++columnIndex) {
                row.put(
                        columnIndex,
                        new TicTacToeFieldImpl(
                                this,
                                TicTacToePosition.of(rowIndex, columnIndex),
                                source.getFieldAtInternal(TicTacToePosition.of(rowIndex, columnIndex)).orElseThrow()
                                        .getState()));
            }
        }
    }

    @Override
    public String toString() {
        final StringBuilder result = new StringBuilder();
        for (int row = 0; row < this.fields.size(); ++row) {
            final Map<Integer, TicTacToeFieldImpl> rowMap = this.fields.get(row);
            for (int column = 0; column < rowMap.size(); ++column) {
                result.append(rowMap.get(column).getState().toString()).append(' ');
            }
            result.append('\n');
        }
        return result.toString();
    }

    @Override
    public boolean equals(final Object obj) {
        if (obj instanceof TicTacToeBoardImpl) {
            final TicTacToeBoardImpl other = (TicTacToeBoardImpl) obj;
            return this.fields.equals(other.fields);
        }
        return false;
    }

    @Override
    public int hashCode() {
        return this.fields.hashCode();
    }

    @Override
    public int getSize() {
        return this.fields.size();
    }

    @Override
    public boolean hasFieldAt(final TicTacToePosition position) {
        final int size = this.getSize();
        final int row = position.getRow();
        final int column = position.getColumn();
        return row >= 0 && row < size && column >= 0 && column < size;
    }

    @Override
    public TicTacToeFieldImpl getFieldAt(final TicTacToePosition position) {
        if (!this.hasFieldAt(position)) {
            throw new IllegalArgumentException(String.format("Position %s out of range.", position));
        }

        return this.getFieldAtInternal(position).orElseThrow();
    }

    @Override
    public List<List<? extends TicTacToeField>> getFields() {
        final List<List<? extends TicTacToeField>> result = new ArrayList<>();
        this.fields.values()
                .forEach((final Map<Integer, TicTacToeFieldImpl> row) -> result.add(new ArrayList<>(row.values())));
        return result;
    }

    @Override
    public Set<TicTacToeRow> getRowsUniformlyMarked() {
        final Set<TicTacToeRow> result = new LinkedHashSet<>();
        for (int rowIndex = 0; rowIndex < this.getSize(); ++rowIndex) {
            addRowIfUniformlyMarked(result, this.getRowOfFields(TicTacToeRowType.ROW, rowIndex));
        }
        for (int columnIndex = 0; columnIndex < this.getSize(); ++columnIndex) {
            addRowIfUniformlyMarked(result, this.getRowOfFields(TicTacToeRowType.COLUMN, columnIndex));
        }
        for (int diagonalIndex = 0; diagonalIndex < 2; ++diagonalIndex) {
            addRowIfUniformlyMarked(result, this.getRowOfFields(TicTacToeRowType.DIAGONAL, diagonalIndex));
        }
        return result;
    }

    @Override
    public Map<TicTacToePosition, TicTacToeFieldImpl> getFieldsBeing(final TicTacToeFieldState fieldState) {
        final Map<TicTacToePosition, TicTacToeFieldImpl> result = new LinkedHashMap<>();
        final int size = this.getSize();
        for (int rowIndex = 0; rowIndex < size; ++rowIndex) {
            for (int columnIndex = 0; columnIndex < size; ++columnIndex) {
                final TicTacToePosition position = TicTacToePosition.of(rowIndex, columnIndex);
                final TicTacToeFieldImpl field = this.getFieldAtInternal(position).orElseThrow();
                if (field.getState().equals(fieldState)) {
                    result.put(position, field);
                }
            }
        }

        return result;
    }

    /**
     * Adds a row to a set of rows if it is uniformly marked, i.e. contains only crosses or only noughts.
     *
     * @param result The result set.
     * @param row    The row to add.
     */
    private static void addRowIfUniformlyMarked(final Set<TicTacToeRow> result, final TicTacToeRow row) {
        final Optional<TicTacToeFieldState> rowState = row.getState();
        if (rowState.isPresent() && !rowState.get().equals(TicTacToeFieldState.EMPTY)) {
            result.add(row);
        }
    }

    @Override
    public TicTacToeBoardImpl deepCopy() {
        return new TicTacToeBoardImpl(this);
    }

    /**
     * Returns a row of fields given the row type and the row index.
     *
     * @param type  The row type.
     * @param index The row index.
     */
    private TicTacToeRow getRowOfFields(final TicTacToeRowType type, final int index) {
        final List<? extends TicTacToeField> fieldsOfRow;

        switch (type) {
        case ROW:
            fieldsOfRow = this.getFields(TicTacToePosition.of(index, 0), TicTacToeDirection.EAST);
            break;
        case COLUMN:
            fieldsOfRow = this.getFields(TicTacToePosition.of(0, index), TicTacToeDirection.SOUTH);
            break;
        case DIAGONAL:
            if (index == 0) {
                fieldsOfRow = this.getFields(TicTacToePosition.of(0, 0), TicTacToeDirection.SOUTHEAST);
            } else {
                fieldsOfRow = this.getFields(TicTacToePosition.of(0, this.getSize() - 1), TicTacToeDirection.SOUTHWEST);
            }
            break;
        default:
            throw new IllegalArgumentException(String.format("Unknown TicTacToeRowType %s.", type));
        }

        return new TicTacToeRowImpl(type, index, fieldsOfRow);
    }

    /**
     * Returns a list of fields starting at a given position and extending towards a given direction till the board's
     * end.
     *
     * @param start     The starting position.
     * @param direction The direction.
     */
    private List<? extends TicTacToeField> getFields(final TicTacToePosition start,
            final TicTacToeDirection direction) {
        final List<TicTacToeField> result = new ArrayList<>();

        TicTacToePosition pos = start;
        Optional<TicTacToeFieldImpl> field = this.getFieldAtInternal(pos);
        while (field.isPresent()) {
            result.add(field.get());
            pos = direction.step(pos);
            field = this.getFieldAtInternal(pos);
        }

        return result;
    }

    /**
     * Returns a field at a given position. If the position does not exist, {@link Optional#empty()} is returned.
     *
     * @param position The position of the field.
     * @return The field.
     */
    private Optional<TicTacToeFieldImpl> getFieldAtInternal(final TicTacToePosition position) {
        return Optional.ofNullable(this.fields.get(position.getRow()))
                .map((final Map<Integer, TicTacToeFieldImpl> map) -> map.get(position.getColumn()));
    }
}
